--[[
BaleLoaderExtension
Specialization for extended bale loader functions.

Author:		Ifko[nator]
Datum:		27.02.2017

Version:	v2.5

History:	v1.0 @ 03.09.2016 - initial implementation
			------------------------------------------------------------------------------------------------------------------
			v1.5 @ 29.09.2016 - MP Fix for change grab distance
			------------------------------------------------------------------------------------------------------------------
			v2.0 @ 06.11.2016 - convert to FS 17 and added some new features
			------------------------------------------------------------------------------------------------------------------
			v2.1 @ 22.11.2016 - MP Fix for Error:

								Warning (LUA): 'streamWriteInt32': Argument 1 has wrong type. Expected: Int. Actual: Nil 
								..\..\..\src\base\scripting\lua\LuaScriptSystem.cpp (604): expectedType == Value::VoidType
								LUA call stack:
								=E:/FS 17/mods/FS17_Arcusin_Autostack_Pack/BaleLoaderExtension.lua (470) : streamWriteInt32
								=dataS/scripts/network/Connection.lua (0) : writeStream
								=E:/FS 17/mods/FS17_Arcusin_Autostack_Pack/BaleLoaderExtension.lua (487) : sendEvent
								=E:/FS 17/mods/FS17_Arcusin_Autostack_Pack/BaleLoaderExtension.lua (330) : sendEvent
								=E:/FS 17/mods/FS17_Arcusin_Autostack_Pack/BaleLoaderExtension.lua (59) : setGrabRange
								=dataS/scripts/vehicles/Vehicle.lua (0)
								=dataS/scripts/vehicles/Vehicle.lua (0) : load
								=dataS/scripts/network/Client.lua (0) : readStream
								=dataS/scripts/network/ConnectionManager.lua (0) : func
								
							  - add function to change rim- and axis color via designConfiguration
							  - deactivate Balecollision in an MP Game - to ensure the Bales will be smoothly unloaded 
			------------------------------------------------------------------------------------------------------------------
			v2.2 @ 23.11.2016 - Fix Error:
								
								=E:/FS 17/mods/FS17_Arcusin_Autostack_Pack/BaleLoaderExtension.lua (284) : setScale
								=dataS/scripts/vehicles/Vehicle.lua (0) : update
								=dataS/scripts/network/Server.lua (0) : update
								=dataS/scripts/BaseMission.lua (0) : update
								=dataS/scripts/FSBaseMission.lua (0) : update
								=dataS/scripts/missions/mission00.lua (0) : update
								=dataS/scripts/main.lua (0) : update
								Warning (LUA): 'setScale': Argument 1 has wrong type. Expected: Int. Actual: Nil
								..\..\..\src\base\scripting\lua\LuaScriptSystem.cpp (604): expectedType == Value::VoidType
			------------------------------------------------------------------------------------------------------------------
			v2.5 @ 27.02.2017 - deactivate particle system for wrapped bales
							  - added new bale grab particle systems for: grass, hay and straw
]]

BaleLoaderExtension = {};
	
function BaleLoaderExtension.prerequisitesPresent(specializations)
	return SpecializationUtil.hasSpecialization(BaleLoader, specializations);
end;

function BaleLoaderExtension:load(savegame)
	self.doStateChange = Utils.appendedFunction(self.doStateChange, BaleLoaderExtension.doStateChangeBaleLoader);
	self.applyDesign = Utils.appendedFunction(self.applyDesign, BaleLoaderExtension.applyDesignVehicle);
	
	self.getAllowsChangeGrabDistance = BaleLoaderExtension.getAllowsChangeGrabDistance;
	self.setGrabMode = BaleLoaderExtension.setGrabMode;
	self.setGrabRange = BaleLoaderExtension.setGrabRange;
	
	if savegame then
		self.grabModeIsProfi = Utils.getNoNil(getXMLBool(savegame.xmlFile, savegame.key .. "#grabModeIsProfi"), true);
	
		if self.grabModeIsProfi then
			self.balePickupRange = 0.23;
		else
			self.balePickupRange = 4;
		end;
	else
		self.grabModeIsProfi = true;
		self.balePickupRange = 0.23;
	end;
	
	self.baleGrabberCheck = Utils.indexToObject(self.components, "0>7|0|0");

	self.baleGrabberRoll = Utils.indexToObject(self.components, "0>7|0|0|2");
	self.baleGrabberRollSmall = Utils.indexToObject(self.components, "0>7|0|0|5");
	
	self.roll1 = Utils.indexToObject(self.components, "0>10|1");
	self.roll2 = Utils.indexToObject(self.components, "0>10|2");
	
	self.baleCollision = Utils.indexToObject(self.components, getXMLString(self.xmlFile, "vehicle.balePlaces#collisisonIndex"));
	
	if not g_currentMission.missionDynamicInfo.isMultiplayer then
		self.collisionMask = getCollisionMask(self.baleCollision);
		
		self.collisionTranslations = {};
		
		for number = 1, #self.balePlaces do
			self.collisionTranslations[number] = Utils.getVectorNFromString(Utils.getNoNil(getXMLString(self.xmlFile, "vehicle.balePlaces.balePlace(" .. tostring(number - 1) .. ")#collisionTranslations"), "0 0.569 5.359"), 3);
			
			--print("collsion(" .. number .. ") value 1 = " .. self.collisionTranslations[number][1] .. " value 2 = " .. self.collisionTranslations[number][2] .. " value 3 = " .. self.collisionTranslations[number][3])
		end;
	end;
	
	if self.isClient then
		self.workSounds = {};
	
		local soundNumber = 0;

		while true do
			local soundKey = "vehicle.workSounds.workSound(" .. tostring(soundNumber) .. ")";
			
			if not hasXMLProperty(self.xmlFile, soundKey) then
				break;
			end;
			
			local workSound = SoundUtil.loadSample(self.xmlFile, {}, soundKey, nil, self.baseDirectory, self.baleGrabberCheck);
			
			if workSound then
				table.insert(self.workSounds, workSound);
			end;
		
			soundNumber = soundNumber + 1;
		end;
	end;
	
	self.baleGrabParticleSystems = {};
	
	self.foundUnsupportetFillType = false;
	
	local particleSystemNumber = 0;
	
	while true do
		local key = string.format("vehicle.baleGrabParticleSystems.baleGrabParticleSystem(%d)", particleSystemNumber);
		
		if not hasXMLProperty(self.xmlFile, key) then
			if self.foundUnsupportetFillType then
				print("[INFO BaleLoaderExtension]:                                \\__ Supportet values: 'straw', 'grass_windrow' or 'dryGrass_windrow'! Using the default value 'straw' instead!");
			end;
			
			break;
		end;
		
		local node = getXMLString(self.xmlFile, key .. "#node");
		node = Utils.indexToObject(self.components, node);
		
		if node ~= nil then
			local baleGrabParticleSystem = {};
			
			baleGrabParticleSystem.particleSystem = {};
			
			baleGrabParticleSystem.disableDuration = Utils.getNoNil(getXMLFloat(self.xmlFile, key .. "#disableDuration"), 0.6) * 1000;
			baleGrabParticleSystem.activeAtBaleGrabbing = Utils.getNoNil(getXMLBool(self.xmlFile, key .. "#activeAtBaleGrabbing"), false);
			baleGrabParticleSystem.onlyActiveAtBaleGrabbing = Utils.getNoNil(getXMLBool(self.xmlFile, key .. "#onlyActiveAtBaleGrabbing"), false);
			baleGrabParticleSystem.activeForFillType = Utils.getNoNil(getXMLString(self.xmlFile, key .. "#activeForFillType"), "straw");
			baleGrabParticleSystem.disableTime = 0;
			
			if baleGrabParticleSystem.activeForFillType ~= "straw" and baleGrabParticleSystem.activeForFillType ~= "grass_windrow" and baleGrabParticleSystem.activeForFillType ~= "dryGrass_windrow" then
				print("[INFO BaleLoaderExtension]: Unsupportet value for 'activeForFillType' ('" .. baleGrabParticleSystem.activeForFillType .. "') in '" .. self.configFileName .. "' ('" .. key .. "')!");

				self.foundUnsupportetFillType = true;
				baleGrabParticleSystem.activeForFillType = "straw";
			end;
			
			ParticleUtil.loadParticleSystem(self.xmlFile, baleGrabParticleSystem.particleSystem, key, self.components, false, nil, self.baseDirectory);
			
			table.insert(self.baleGrabParticleSystems, baleGrabParticleSystem);
		end;
		
		particleSystemNumber = particleSystemNumber + 1;
	end;
	
	self.baleloaderTurnedOnScrollersModdingWelt = Utils.loadScrollers(self.components, self.xmlFile, "vehicle.baleloaderTurnedOnScrollersModdingWelt.baleloaderTurnedOnScrollerModdingWelt", {}, false);
end;

function BaleLoaderExtension:applyDesignVehicle(xmlFile, configDesignId)
	local designKey = string.format("vehicle.designConfigurations.designConfiguration(%d)", configDesignId-1);
    
	if not hasXMLProperty(xmlFile, designKey) then
        print("Warning: Invalid design configuration " .. configDesignId);
        
		return;
    end;
    
	local axisColor = getXMLString(xmlFile, designKey .. "#axisColor");
	local rimColor = getXMLString(xmlFile, designKey .. "#rimColor");
    
	for _, wheel in pairs(self.wheels) do
		if axisColor then
			self.axisColor = Vehicle.getColorFromString(axisColor);
			
			if wheel.wheelHub then
				local r, g, b, a = unpack(self.axisColor);
            
				setShaderParameter(wheel.wheelHub, "colorScale", r, g, b, a, false);
			
				--print("Change axis color to: " .. axisColor);
			end;
		end;
		
		if rimColor then
			self.rimColor = Vehicle.getColorFromString(rimColor);
			
			local r, g, b, a = unpack(self.rimColor);
        
			if wheel.wheelOuterRim then
				setShaderParameter(wheel.wheelOuterRim, "colorScale", r, g, b, a, false);
			
				--print("Change outer rim color to: " .. rimColor);
			end;
			
			if wheel.wheelInnerRim then
				setShaderParameter(wheel.wheelInnerRim, "colorScale", r, g, b, a, false);
			
				--print("Change inner rim color to: " .. rimColor);
			end;
		end;
	end;
end;

function BaleLoaderExtension:doStateChangeBaleLoader(id, nearestBaleServerId)
	if id == BaleLoader.CHANGE_GRAB_BALE then
		--## play animation for pick up the bale to the bale grabber.
		self:playAnimation("baleGrabberLoadBale", 1, nil, true);
	elseif id == BaleLoader.CHANGE_GRAB_MOVE_DOWN then
		--## reset bale animation to start values for the next bale!
		self:playAnimation("baleGrabberLoadBale", -5000, nil, true);
	end;
end;

function BaleLoaderExtension:getAllowsChangeGrabDistance()
	return self.emptyState == BaleLoader.EMPTY_NONE and self.isInWorkPosition;
end;

function BaleLoaderExtension:update(dt)
	if self:getIsActive() then
		local grabberRotX, grabberRotY, _ = getRotation(self.baleGrabberCheck);
		local _, baleTransY, baleTransZ = getTranslation(self.baleGrabber.grabNode);
		
		local rollRotX, _, _ = getRotation(self.baleGrabberRoll);
		local _, _, rollRotZ = getRotation(self.roll1);
		
		local speed = 0.012 * dt;
		
		grabberRotX, grabberRotY, grabberRotZ = Utils.round(math.deg(grabberRotX), 2), Utils.round(math.deg(grabberRotY), 2);
		baleTransY, baleTransZ = Utils.round(baleTransY, 2), Utils.round(baleTransZ, 2);
		
		local isPickUpBale = baleTransY ~= 0.18 and baleTransZ ~= 0.8 and self.grabberMoveState == BaleLoader.GRAB_MOVE_UP;
		local isLoadingBaleToPlatform = grabberRotX == -35 and grabberRotY == -90 and self.grabberMoveState == BaleLoader.GRAB_DROP_BALE;
		
		local baleWrappingState = 0;
		local baleFillType = "unknown";
		
		if self.baleGrabber.currentBale then
			local bale = networkGetObject(self.baleGrabber.currentBale);
			
			if bale then
				baleWrappingState = Utils.getNoNil(bale.wrappingState, 0);
				baleFillType = Utils.getNoNil(FillUtil.fillTypeIntToName[bale.fillType], "unknown");
			end;
		end;
		
		--setTextAlignment(RenderText.ALIGN_CENTER);
		--renderText(0.5, 0.5, 0.02, "current bale fill type at the bale grabber = " .. baleFillType);
		
		if isPickUpBale then
			rollRotX = rollRotX - speed;
			
			if self.baleGrabParticleSystems then
				for _, baleGrabParticleSystem in pairs(self.baleGrabParticleSystems) do
					if (baleGrabParticleSystem.activeAtBaleGrabbing or baleGrabParticleSystem.onlyActiveAtBaleGrabbing) and baleWrappingState == 0 and baleFillType == baleGrabParticleSystem.activeForFillType	then
						ParticleUtil.setEmittingState(baleGrabParticleSystem.particleSystem, true);
						
						baleGrabParticleSystem.disableTime = g_currentMission.time + baleGrabParticleSystem.disableDuration;
					end;
				end;
			end;
		elseif isLoadingBaleToPlatform then
			rollRotX = rollRotX - speed;
			rollRotZ = rollRotZ - speed;
			
			if self.baleGrabParticleSystems then
				for _, baleGrabParticleSystem in pairs(self.baleGrabParticleSystems) do
					if not baleGrabParticleSystem.onlyActiveAtBaleGrabbing and baleWrappingState == 0 and baleFillType == baleGrabParticleSystem.activeForFillType then	
						ParticleUtil.setEmittingState(baleGrabParticleSystem.particleSystem, true);
					
						baleGrabParticleSystem.disableTime = g_currentMission.time + baleGrabParticleSystem.disableDuration;
					end;
				end;
			end;
		end;
		
		if isPickUpBale or isLoadingBaleToPlatform then
			if self.workSounds then
				for _, workSound in pairs(self.workSounds) do	
					if self:getIsActiveForSound() then	
						SoundUtil.playSample(workSound, 0, 0, nil);
					else --if self.isControlled then
						SoundUtil.play3DSample(workSound);
					end;
				end;
			end
		else
			if self.workSounds then
				for _, workSound in pairs(self.workSounds) do
					SoundUtil.stopSample(workSound, true);
					SoundUtil.stop3DSample(workSound);
				end;
			end;
			
			if self.isClient and self.baleGrabParticleSystems then 
				for _, baleGrabParticleSystem in pairs(self.baleGrabParticleSystems) do
					if baleGrabParticleSystem.disableTime ~= 0 and baleGrabParticleSystem.disableTime < g_currentMission.time then
						ParticleUtil.setEmittingState(baleGrabParticleSystem.particleSystem, false);
					
						baleGrabParticleSystem.disableTime = 0;
					end;
				end;
			end;
		end;
		
		Utils.updateScrollers(self.baleloaderTurnedOnScrollersModdingWelt, dt, isPickUpBale or isLoadingBaleToPlatform);
		
		setRotation(self.baleGrabberRoll, rollRotX, 0, 0);
		setRotation(self.baleGrabberRollSmall, rollRotX, 0, 0);
		
		setRotation(self.roll1, 0, 0, rollRotZ);
		setRotation(self.roll2, 0, 0, rollRotZ);
		
		if not g_currentMission.missionDynamicInfo.isMultiplayer then
			local scaleX, scaleY, scaleZ = 1, 1, 1;
			local baleRows = 1;
			
			if self:getFillLevel() == 0 then
				scaleX, scaleY, scaleZ = 0, 0, 0;
			elseif self:getFillLevel() >= 2 then
				scaleY = scaleY * 2;
			end;
			
			if self:getFillLevel() >= 3 and self:getFillLevel() < 5 then
				baleRows = 2;
			elseif self:getFillLevel() >= 5 and self:getFillLevel() < 7 then
				baleRows = 3;
			elseif self:getFillLevel() >= 7 and self:getFillLevel() < 9 then
				baleRows = 4;
			elseif self:getFillLevel() >= 9 and self:getFillLevel() < 11 then
				baleRows = 5;
			elseif self:getFillLevel() >= 11 and self:getFillLevel() < 13 then
				baleRows = 6;
			elseif self:getFillLevel() >= 13 and self:getFillLevel() < 15 then
				baleRows = 7;
			elseif self:getFillLevel() >= 15 then
				baleRows = 8;
			end;
			
			setScale(self.baleCollision, scaleX, scaleY, scaleZ * baleRows);
			
			if self:getFillLevel() == 0 then	
				baleRows = #self.balePlaces;
			end;
			
			if self.emptyState == BaleLoader.EMPTY_ROTATE_PLATFORM or self.emptyState == BaleLoader.EMPTY_SINK then
				setTranslation(self.baleCollision, self.collisionTranslations[baleRows][1], self.collisionTranslations[baleRows][2], self.collisionTranslations[baleRows][3]);
			end;
			
			if self.emptyState == BaleLoader.EMPTY_WAIT_TO_SINK then
				setCollisionMask(self.baleCollision, 0);
			elseif self.emptyState == BaleLoader.EMPTY_NONE then
				setCollisionMask(self.baleCollision, self.collisionMask);
			end;
		else
			local scaleX, scaleY, scaleZ = getScale(self.baleCollision);
			
			if (scaleX and scaleY and scaleZ) ~= 0 then
				setScale(self.baleCollision, 0, 0, 0);
			end;
		end;
		
		--[[
		renderText(0.5, 0.52, 0.02, "translationX = '" .. self.collisionTranslations[baleRows][1] .. "' translationY = '" .. self.collisionTranslations[baleRows][2] .. "' translationZ = '" .. self.collisionTranslations[baleRows][3] .. "'.");
		renderText(0.5, 0.5, 0.02, "scaleX = '" .. scaleX .. "' scaleY = '" .. scaleY .. "' scaleZ = '" .. scaleZ * baleRows .. "'.");
		renderText(0.5, 0.48, 0.02, "collision mask from the bale collision = " .. getCollisionMask(self.baleCollision));
		renderText(0.5, 0.46, 0.02, "number of collison translations = " .. #self.collisionTranslations);
		]]
	
		if self:getFillLevel() == self:getCapacity()
			and self.emptyState == BaleLoader.EMPTY_NONE 
			and self.grabberMoveState == nil 
			and self.rotatePlatformDirection == 0 
			and self.isInWorkPosition 
		then
		--[[baleloader is full, 
				baleloader state is 'none', 
				grabber is not moving,
				stackplatform is set down,
				and grabber is in work position 
			- 
				move grabber to transport position
		]]
			g_client:getServerConnection():sendEvent(BaleLoaderStateEvent:new(self, BaleLoader.CHANGE_BUTTON_WORK_TRANSPORT));
		end;
	end;

	if self:getIsActiveForInput() and self:getAllowsChangeGrabDistance() then
		if InputBinding.hasEvent(InputBinding.IMPLEMENT_EXTRA2) then
			self:setGrabMode(not self.grabModeIsProfi, false);
			
			if self.grabModeIsProfi then
				self:setGrabRange(0.23, false);
			else
				self:setGrabRange(4, false);
			end;
		end;
	end;
	
	
end;

function BaleLoaderExtension:setGrabMode(state, noEventSend)
	if state ~= self.grabModeIsProfi then
		self.grabModeIsProfi = state;
		BaleLoaderExtensionGrabModeEvent.sendEvent(self, state, noEventSend);
	end;
end;

function BaleLoaderExtension:setGrabRange(range, noEventSend)
	if range ~= self.balePickupRange then
		self.balePickupRange = range;
		BaleLoaderExtensionGrabRangeEvent.sendEvent(self, range, noEventSend);
	end;
end;

function BaleLoaderExtension:draw()
	if self:getIsActiveForInput() and self:getAllowsChangeGrabDistance() then
		g_currentMission:addHelpButtonText(g_i18n:getText("CANGE_GRAB_MODE"), InputBinding.IMPLEMENT_EXTRA2);
	
		if self.grabModeIsProfi then
			g_currentMission:addExtraPrintText(g_i18n:getText("PROFI_GRAB_MODE"));
		else	
			g_currentMission:addExtraPrintText(g_i18n:getText("ROOKIE_GRAB_MODE"));
		end;
	end;
end;

function BaleLoaderExtension:getSaveAttributesAndNodes(nodeIdent)
	local attributes = 'grabModeIsProfi="' .. tostring(self.grabModeIsProfi) .. '"';
	
	return attributes;
end;

function BaleLoaderExtension:delete()
	if self.isClient and self.workSounds then
		for _, workSound in pairs(self.workSounds) do
			SoundUtil.deleteSample(workSound);
		end;
	end;
	
	if self.baleGrabParticleSystems then
        for _, baleGrabParticleSystem in pairs(self.baleGrabParticleSystems) do
			ParticleUtil.deleteParticleSystem(baleGrabParticleSystem.particleSystem);
		end;
	end;
end;

function BaleLoaderExtension:onDeactivate()
	if self.isClient and self.workSounds then
		for _, workSound in pairs(self.workSounds) do	
			SoundUtil.stopSample(workSound, true);
			SoundUtil.stop3DSample(workSound);
		end;
    end;
	
	if self.baleGrabParticleSystems then
        for _, baleGrabParticleSystem in pairs(self.baleGrabParticleSystems) do
			ParticleUtil.setEmittingState(baleGrabParticleSystem.particleSystem, false);
		end;
	end;
end;

--## unused but needed functions

function BaleLoaderExtension:mouseEvent(posX, posY, isDown, isUp, button)end;
function BaleLoaderExtension:keyEvent(unicode, sym, modifier, isDown)end;

--## MP Stuff

--## set grab mode

BaleLoaderExtensionGrabModeEvent = {};
BaleLoaderExtensionGrabModeEvent_mt = Class(BaleLoaderExtensionGrabModeEvent, Event);

InitEventClass(BaleLoaderExtensionGrabModeEvent, "BaleLoaderExtensionGrabModeEvent");

function BaleLoaderExtensionGrabModeEvent:emptyNew()
	local self = Event:new(BaleLoaderExtensionGrabModeEvent_mt);
    
	return self;
end;

function BaleLoaderExtensionGrabModeEvent:new(object, state)
	local self = BaleLoaderExtensionGrabModeEvent:emptyNew();
	
	self.object = object;
	self.state = state;
	
	return self;
end;

function BaleLoaderExtensionGrabModeEvent:readStream(streamId, connection)
	local id = streamReadInt32(streamId);
	
	self.object = networkGetObject(id);
	self.state = streamReadBool(streamId);
    self:run(connection);
end;

function BaleLoaderExtensionGrabModeEvent:writeStream(streamId, connection)
	streamWriteInt32(streamId, networkGetObjectId(self.object));
	streamWriteBool(streamId, self.state);
end;

function BaleLoaderExtensionGrabModeEvent:run(connection)
	self.object:setGrabMode(self.state, true);
	
	if not connection:getIsServer() then
		g_server:broadcastEvent(BaleLoaderExtensionGrabModeEvent:new(self.object, self.state), nil, connection, self.object);
	end;
end;

function BaleLoaderExtensionGrabModeEvent.sendEvent(object, state, noEventSend)
	if noEventSend == nil or noEventSend == false then
		if g_server then
			g_server:broadcastEvent(BaleLoaderExtensionGrabModeEvent:new(object, state), nil, nil, object);
		else
			g_client:getServerConnection():sendEvent(BaleLoaderExtensionGrabModeEvent:new(object, state));
		end;
	end;
end;

-----------------------------------------------------------------------------------------------------------------------------------

--## set grab distance

BaleLoaderExtensionGrabRangeEvent = {};
BaleLoaderExtensionGrabRangeEvent_mt = Class(BaleLoaderExtensionGrabRangeEvent, Event);

InitEventClass(BaleLoaderExtensionGrabRangeEvent, "BaleLoaderExtensionGrabRangeEvent");

function BaleLoaderExtensionGrabRangeEvent:emptyNew()
	local self = Event:new(BaleLoaderExtensionGrabRangeEvent_mt);
   
	return self;
end;

function BaleLoaderExtensionGrabRangeEvent:new(object, range)
	local self = BaleLoaderExtensionGrabRangeEvent:emptyNew();
	
	self.object = object;
	self.range = range;
	
	return self;
end;

function BaleLoaderExtensionGrabRangeEvent:readStream(streamId, connection)
	local id = streamReadInt32(streamId);
	
	self.object = networkGetObject(id);
	self.range = streamReadFloat32(streamId);
    self:run(connection);
end;

function BaleLoaderExtensionGrabRangeEvent:writeStream(streamId, connection)
	streamWriteInt32(streamId, networkGetObjectId(self.object));
	streamWriteFloat32(streamId, self.range);
end;

function BaleLoaderExtensionGrabRangeEvent:run(connection)
	self.object:setGrabRange(self.range, true);
	
	if not connection:getIsServer() then
		g_server:broadcastEvent(BaleLoaderExtensionGrabRangeEvent:new(self.object, self.range), nil, connection, self.object);
	end;
end;

function BaleLoaderExtensionGrabRangeEvent.sendEvent(object, range, noEventSend)
	if noEventSend == nil or noEventSend == false then
		if g_server then
			g_server:broadcastEvent(BaleLoaderExtensionGrabRangeEvent:new(object, range), nil, nil, object);
		else
			g_client:getServerConnection():sendEvent(BaleLoaderExtensionGrabRangeEvent:new(object, range));
		end;
	end;
end;